# -*- coding:utf-8 -*-

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110- 1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# <pep8 compliant>

# ----------------------------------------------------------
# Author: Stephen Leger (s-leger)
#
# ----------------------------------------------------------

from math import cos, sin, tan, sqrt, atan2, pi
from mathutils import Vector


class Panel():
    """
        Define a bevel profil
        index: array associate each y with a coord circle and a x
        x = array of x of unique points in the profil relative to origin (0, 0) is bottom left
        y = array of y of all points in the profil relative to origin (0, 0) is bottom left
        idmat = array of material index for each segment
        when path is not closed, start and end caps are generated

        shape is the loft profile
        path is the loft path

        Open shape:

        x = [0,1]
        y = [0,1,1, 0]
        index = [0, 0,1,1]
        closed_shape = False

        1 ____2
         |    |
         |    |
         |    |
        0     3

        Closed shape:

        x = [0,1]
        y = [0,1,1, 0]
        index = [0, 0,1,1]
        closed_shape = True

        1 ____2
         |    |
         |    |
         |____|
        0     3

        Side Caps (like glass for window):

        x = [0,1]
        y = [0,1,1, 0.75, 0.25, 0]
        index = [0, 0,1,1,1,1]
        closed_shape = True
        side_caps = [3,4]

        1 ____2        ____
         |   3|__cap__|    |
         |   4|_______|    |
         |____|       |____|
        0     5

    """
    def __init__(self, closed_shape, index, x, y, idmat, side_cap_front=-1, side_cap_back=-1, closed_path=True,
            subdiv_x=0, subdiv_y=0, user_path_verts=0, user_path_uv_v=None):

        self.closed_shape = closed_shape
        self.closed_path = closed_path
        self.index = index
        self.x = x
        self.y = y
        self.idmat = idmat
        self.side_cap_front = side_cap_front
        self.side_cap_back = side_cap_back
        self.subdiv_x = subdiv_x
        self.subdiv_y = subdiv_y
        self.user_path_verts = user_path_verts
        self.user_path_uv_v = user_path_uv_v

    @property
    def n_pts(self):
        return len(self.y)

    @property
    def profil_faces(self):
        """
            number of faces for each section
        """
        if self.closed_shape:
            return len(self.y)
        else:
            return len(self.y) - 1

    @property
    def uv_u(self):
        """
            uvs of profil (absolute value)
        """
        x = [self.x[i] for i in self.index]
        x.append(x[0])
        y = [y for y in self.y]
        y.append(y[0])
        uv_u = []
        uv = 0
        uv_u.append(uv)
        for i in range(len(self.index)):
            dx = x[i + 1] - x[i]
            dy = y[i + 1] - y[i]
            uv += sqrt(dx * dx + dy * dy)
            uv_u.append(uv)
        return uv_u

    def path_sections(self, steps, path_type):
        """
            number of verts and faces sections along path
        """
        n_path_verts = 2
        if path_type in ['QUADRI', 'RECTANGLE']:
            n_path_verts = 4 + self.subdiv_x + 2 * self.subdiv_y
            if self.closed_path:
                n_path_verts += self.subdiv_x
        elif path_type in ['ROUND', 'ELLIPSIS']:
            n_path_verts = steps + 3
        elif path_type == 'CIRCLE':
            n_path_verts = steps
        elif path_type == 'TRIANGLE':
            n_path_verts = 3
        elif path_type == 'PENTAGON':
            n_path_verts = 5
        elif path_type == 'USER_DEFINED':
            n_path_verts = self.user_path_verts
        if self.closed_path:
            n_path_faces = n_path_verts
        else:
            n_path_faces = n_path_verts - 1
        return n_path_verts, n_path_faces

    def n_verts(self, steps, path_type):
        n_path_verts, n_path_faces = self.path_sections(steps, path_type)
        return self.n_pts * n_path_verts

    ############################
    # Geomerty
    ############################

    def _intersect_line(self, center, basis, x):
        """ upper intersection of line parallel to y axis and a triangle
            where line is given by x origin
            top by center, basis size as float
            return float y of upper intersection point

            center.x and center.y are absolute
            a 0 center.x lie on half size
            a 0 center.y lie on basis
        """
        if center.x > 0:
            dx = x - center.x
        else:
            dx = center.x - x
        p = center.y / basis
        return center.y + dx * p

    def _intersect_triangle(self, center, basis, x):
        """ upper intersection of line parallel to y axis and a triangle
            where line is given by x origin
            top by center, basis size as float
            return float y of upper intersection point

            center.x and center.y are absolute
            a 0 center.x lie on half size
            a 0 center.y lie on basis
        """
        if x > center.x:
            dx = center.x - x
            sx = 0.5 * basis - center.x
        else:
            dx = x - center.x
            sx = 0.5 * basis + center.x
        if sx == 0:
            sx = basis
        p = center.y / sx
        return center.y + dx * p

    def _intersect_circle(self, center, radius, x):
        """ upper intersection of line parallel to y axis and a circle
            where line is given by x origin
            circle by center, radius as float
            return float y of upper intersection point, float angle
        """
        dx = x - center.x
        d = (radius * radius) - (dx * dx)
        if d <= 0:
            if x > center.x:
                return center.y, 0
            else:
                return center.y, pi
        else:
            y = sqrt(d)
            return center.y + y, atan2(y, dx)

    def _intersect_elipsis(self, center, radius, x):
        """ upper intersection of line parallel to y axis and an ellipsis
            where line is given by x origin
            circle by center, radius.x and radius.y semimajor and seminimor axis (half width and height) as float
            return float y of upper intersection point, float angle
        """
        dx = x - center.x
        d2 = dx * dx
        A = 1 / radius.y / radius.y
        C = d2 / radius.x / radius.x - 1
        d = - 4 * A * C
        if d <= 0:
            if x > center.x:
                return center.y, 0
            else:
                return center.y, pi
        else:
            y0 = sqrt(d) / 2 / A
            d = (radius.x * radius.x) - d2
            y = sqrt(d)
            return center.y + y0, atan2(y, dx)

    def _intersect_arc(self, center, radius, x_left, x_right):
        y0, a0 = self._intersect_circle(center, radius.x, x_left)
        y1, a1 = self._intersect_circle(center, radius.x, x_right)
        da = (a1 - a0)
        if da < -pi:
            da += 2 * pi
        if da > pi:
            da -= 2 * pi
        return y0, y1, a0, da

    def _intersect_arc_elliptic(self, center, radius, x_left, x_right):
        y0, a0 = self._intersect_elipsis(center, radius, x_left)
        y1, a1 = self._intersect_elipsis(center, radius, x_right)
        da = (a1 - a0)
        if da < -pi:
            da += 2 * pi
        if da > pi:
            da -= 2 * pi
        return y0, y1, a0, da

    def _get_ellispe_coords(self, steps, offset, center, origin, size, radius, x, pivot, bottom_y=0):
        """
            Rectangle with single arc on top
        """
        x_left = size.x / 2 * (pivot - 1) + x
        x_right = size.x / 2 * (pivot + 1) - x
        cx = center.x - origin.x
        cy = offset.y + center.y - origin.y
        y0, y1, a0, da = self._intersect_arc_elliptic(center, radius, origin.x + x_left, origin.x + x_right)
        da /= steps
        coords = []
        # bottom left
        if self.closed_path:
            coords.append((offset.x + x_left, offset.y + x + bottom_y))
        else:
            coords.append((offset.x + x_left, offset.y + bottom_y))
        # top left
        coords.append((offset.x + x_left, offset.y + y0 - origin.y))
        for i in range(1, steps):
            a = a0 + i * da
            coords.append((offset.x + cx + cos(a) * radius.x, cy + sin(a) * radius.y))
        # top right
        coords.append((offset.x + x_right, offset.y + y1 - origin.y))
        # bottom right
        if self.closed_path:
            coords.append((offset.x + x_right, offset.y + x + bottom_y))
        else:
            coords.append((offset.x + x_right, offset.y + bottom_y))
        return coords

    def _get_arc_coords(self, steps, offset, center, origin, size, radius, x, pivot, bottom_y=0):
        """
            Rectangle with single arc on top
        """
        x_left = size.x / 2 * (pivot - 1) + x
        x_right = size.x / 2 * (pivot + 1) - x
        cx = offset.x + center.x - origin.x
        cy = offset.y + center.y - origin.y
        y0, y1, a0, da = self._intersect_arc(center, radius, origin.x + x_left, origin.x + x_right)
        da /= steps
        coords = []

        # bottom left
        if self.closed_path:
            coords.append((offset.x + x_left, offset.y + x + bottom_y))
        else:
            coords.append((offset.x + x_left, offset.y + bottom_y))

        # top left
        coords.append((offset.x + x_left, offset.y + y0 - origin.y))

        for i in range(1, steps):
            a = a0 + i * da
            coords.append((cx + cos(a) * radius.x, cy + sin(a) * radius.x))

        # top right
        coords.append((offset.x + x_right, offset.y + y1 - origin.y))

        # bottom right
        if self.closed_path:
            coords.append((offset.x + x_right, offset.y + x + bottom_y))
        else:
            coords.append((offset.x + x_right, offset.y + bottom_y))

        return coords

    def _get_circle_coords(self, steps, offset, center, origin, radius):
        """
            Full circle
        """
        cx = offset.x + center.x - origin.x
        cy = offset.y + center.y - origin.y
        a = -2 * pi / steps
        return [(cx + cos(i * a) * radius.x, cy + sin(i * a) * radius.x) for i in range(steps)]

    def _get_rectangular_coords(self, offset, size, x, pivot, bottom_y=0):
        coords = []

        x_left = offset.x + size.x / 2 * (pivot - 1) + x
        x_right = offset.x + size.x / 2 * (pivot + 1) - x

        if self.closed_path:
            y0 = offset.y + x + bottom_y
        else:
            y0 = offset.y + bottom_y
        y1 = offset.y + size.y - x

        dy = (y1 - y0) / (1 + self.subdiv_y)
        dx = (x_right - x_left) / (1 + self.subdiv_x)

        # bottom left
        # coords.append((x_left, y0))

        # subdiv left
        for i in range(self.subdiv_y + 1):
            coords.append((x_left, y0 + i * dy))

        # top left
        # coords.append((x_left, y1))

        # subdiv top
        for i in range(self.subdiv_x + 1):
            coords.append((x_left + dx * i, y1))

        # top right
        # coords.append((x_right, y1))
        # subdiv right
        for i in range(self.subdiv_y + 1):
            coords.append((x_right, y1 - i * dy))

        # subdiv bottom
        if self.closed_path:
            for i in range(self.subdiv_x + 1):
                coords.append((x_right - dx * i, y0))
        else:
            # bottom right
            coords.append((x_right, y0))

        return coords

    def _get_vertical_rectangular_trapezoid_coords(self, offset, center, origin, size, basis, x, pivot, bottom_y=0):
        """
            Rectangular trapezoid vertical
            basis is the full width of a triangular area the trapezoid lie into
            center.y is the height of triagular area from top
            center.x is the offset from basis center

            |\
            | \
            |__|
        """
        coords = []
        x_left = size.x / 2 * (pivot - 1) + x
        x_right = size.x / 2 * (pivot + 1) - x
        sx = x * sqrt(basis * basis + center.y * center.y) / basis
        dy = size.y + offset.y - sx
        y0 = self._intersect_line(center, basis, origin.x + x_left)
        y1 = self._intersect_line(center, basis, origin.x + x_right)
        # bottom left
        if self.closed_path:
            coords.append((offset.x + x_left, offset.y + x + bottom_y))
        else:
            coords.append((offset.x + x_left, offset.y + bottom_y))
        # top left
        coords.append((offset.x + x_left, dy - y0))
        # top right
        coords.append((offset.x + x_right, dy - y1))
        # bottom right
        if self.closed_path:
            coords.append((offset.x + x_right, offset.y + x + bottom_y))
        else:
            coords.append((offset.x + x_right, offset.y + bottom_y))
        return coords

    def _get_horizontal_rectangular_trapezoid_coords(self, offset, center, origin, size, basis, x, pivot, bottom_y=0):
        """
            Rectangular trapeze horizontal
            basis is the full width of a triangular area the trapezoid lie into
            center.y is the height of triagular area from top to basis
            center.x is the offset from basis center
             ___
            |   \
            |____\

            TODO: correct implementation
        """
        raise NotImplementedError

    def _get_pentagon_coords(self, offset, center, origin, size, basis, x, pivot, bottom_y=0):
        """
            TODO: correct implementation
                /\
               /  \
              |    |
              |____|
        """
        raise NotImplementedError

    def _get_triangle_coords(self, offset, center, origin, size, basis, x, pivot, bottom_y=0):
        coords = []
        x_left = offset.x + size.x / 2 * (pivot - 1) + x
        x_right = offset.x + size.x / 2 * (pivot + 1) - x

        # bottom left
        if self.closed_path:
            coords.append((x_left, offset.y + x + bottom_y))
        else:
            coords.append((x_left, offset.y + bottom_y))
        # top center
        coords.append((center.x, offset.y + center.y))
        # bottom right
        if self.closed_path:
            coords.append((x_right, offset.y + x + bottom_y))
        else:
            coords.append((x_right, offset.y + bottom_y))
        return coords

    def _get_horizontal_coords(self, offset, size, x, pivot):
        coords = []
        x_left = offset.x + size.x / 2 * (pivot - 1)
        x_right = offset.x + size.x / 2 * (pivot + 1)
        # left
        coords.append((x_left, offset.y + x))
        # right
        coords.append((x_right, offset.y + x))
        return coords

    def _get_vertical_coords(self, offset, size, x, pivot):
        coords = []
        x_left = offset.x + size.x / 2 * (pivot - 1) + x
        # top
        coords.append((x_left, offset.y + size.y))
        # bottom
        coords.append((x_left, offset.y))
        return coords

    def choose_a_shape_in_tri(self, center, origin, size, basis, pivot):
        """
            Choose which shape inside either a tri or a pentagon
        """
        cx = (0.5 * basis + center.x) - origin.x
        cy = center.y - origin.y
        x_left = size.x / 2 * (pivot - 1)
        x_right = size.x / 2 * (pivot + 1)
        y0 = self.intersect_triangle(cx, cy, basis, x_left)
        y1 = self.intersect_triangle(cx, cy, basis, x_right)
        if (y0 == 0 and y1 == 0) or ((y0 == 0 or y1 == 0) and (y0 == cy or y1 == cy)):
            return 'TRIANGLE'
        elif x_right <= cx or x_left >= cx:
            # single side of triangle
            # may be horizontal or vertical rectangular trapezoid
            # horizontal if size.y < center.y
            return 'QUADRI'
        else:
            # both sides of triangle
            # may be horizontal trapezoid or pentagon
            # horizontal trapezoid if size.y < center.y
            return 'PENTAGON'

    ############################
    # Vertices
    ############################

    def vertices(self, steps, offset, center, origin, size, radius,
            angle_y, pivot, shape_z=None, path_type='ROUND', axis='XZ'):

        verts = []
        if shape_z is None:
            shape_z = [0 for x in self.x]
        if path_type == 'ROUND':
            coords = [self._get_arc_coords(steps, offset, center, origin,
                size, Vector((radius.x - x, 0)), x, pivot, shape_z[i]) for i, x in enumerate(self.x)]
        elif path_type == 'ELLIPSIS':
            coords = [self._get_ellispe_coords(steps, offset, center, origin,
                size, Vector((radius.x - x, radius.y - x)), x, pivot, shape_z[i]) for i, x in enumerate(self.x)]
        elif path_type == 'QUADRI':
            coords = [self._get_vertical_rectangular_trapezoid_coords(offset, center, origin,
                size, radius.x, x, pivot) for i, x in enumerate(self.x)]
        elif path_type == 'HORIZONTAL':
            coords = [self._get_horizontal_coords(offset, size, x, pivot)
                for i, x in enumerate(self.x)]
        elif path_type == 'VERTICAL':
            coords = [self._get_vertical_coords(offset, size, x, pivot)
                for i, x in enumerate(self.x)]
        elif path_type == 'CIRCLE':
            coords = [self._get_circle_coords(steps, offset, center, origin, Vector((radius.x - x, 0)))
                for i, x in enumerate(self.x)]
        else:
            coords = [self._get_rectangular_coords(offset, size, x, pivot, shape_z[i])
                for i, x in enumerate(self.x)]
        # vertical panel (as for windows)
        if axis == 'XZ':
            for i in range(len(coords[0])):
                for j, p in enumerate(self.index):
                    x, z = coords[p][i]
                    y = self.y[j]
                    verts.append((x, y, z))
        # horizontal panel (table and so on)
        elif axis == 'XY':
            for i in range(len(coords[0])):
                for j, p in enumerate(self.index):
                    x, y = coords[p][i]
                    z = self.y[j]
                    verts.append((x, y, z))
        return verts

    ############################
    # Faces
    ############################

    def _faces_cap(self, faces, n_path_verts, offset):
        if self.closed_shape and not self.closed_path:
            last_point = offset + self.n_pts * n_path_verts - 1
            faces.append(tuple([offset + i for i in range(self.n_pts)]))
            faces.append(tuple([last_point - i for i in range(self.n_pts)]))

    def _faces_closed(self, n_path_faces, offset):
        faces = []
        n_pts = self.n_pts
        for i in range(n_path_faces):
            k0 = offset + i * n_pts
            if self.closed_path and i == n_path_faces - 1:
                k1 = offset
            else:
                k1 = k0 + n_pts
            for j in range(n_pts - 1):
                faces.append((k1 + j, k1 + j + 1, k0 + j + 1, k0 + j))
            # close profile
            faces.append((k1 + n_pts - 1, k1, k0, k0 + n_pts - 1))
        return faces

    def _faces_open(self, n_path_faces, offset):
        faces = []
        n_pts = self.n_pts
        for i in range(n_path_faces):
            k0 = offset + i * n_pts
            if self.closed_path and i == n_path_faces - 1:
                k1 = offset
            else:
                k1 = k0 + n_pts
            for j in range(n_pts - 1):
                faces.append((k1 + j, k1 + j + 1, k0 + j + 1, k0 + j))
        return faces

    def _faces_side(self, faces, n_path_verts, start, reverse, offset):
        n_pts = self.n_pts
        vf = [offset + start + n_pts * f for f in range(n_path_verts)]
        if reverse:
            faces.append(tuple(reversed(vf)))
        else:
            faces.append(tuple(vf))

    def faces(self, steps, offset=0, path_type='ROUND'):
        n_path_verts, n_path_faces = self.path_sections(steps, path_type)
        if self.closed_shape:
            faces = self._faces_closed(n_path_faces, offset)
        else:
            faces = self._faces_open(n_path_faces, offset)
        if self.side_cap_front > -1:
            self._faces_side(faces, n_path_verts, self.side_cap_front, False, offset)
        if self.side_cap_back > -1:
            self._faces_side(faces, n_path_verts, self.side_cap_back, True, offset)
        self._faces_cap(faces, n_path_verts, offset)
        return faces

    ############################
    # Uvmaps
    ############################

    def uv(self, steps, center, origin, size, radius, angle_y, pivot, x, x_cap, path_type='ROUND'):
        uvs = []
        n_path_verts, n_path_faces = self.path_sections(steps, path_type)
        if path_type in ['ROUND', 'ELLIPSIS']:
            x_left = size.x / 2 * (pivot - 1) + x
            x_right = size.x / 2 * (pivot + 1) - x
            if path_type == 'ELLIPSIS':
                y0, y1, a0, da = self._intersect_arc_elliptic(center, radius, x_left, x_right)
            else:
                y0, y1, a0, da = self._intersect_arc(center, radius, x_left, x_right)
            uv_r = abs(da) * radius.x / steps
            uv_v = [uv_r for i in range(steps)]
            uv_v.insert(0, y0 - origin.y)
            uv_v.append(y1 - origin.y)
            uv_v.append(size.x)
        elif path_type == 'USER_DEFINED':
            uv_v = self.user_path_uv_v
        elif path_type == 'CIRCLE':
            uv_r = 2 * pi * radius.x / steps
            uv_v = [uv_r for i in range(steps + 1)]
        elif path_type == 'QUADRI':
            dy = 0.5 * tan(angle_y) * size.x
            uv_v = [size.y - dy, size.x, size.y + dy, size.x]
        elif path_type == 'HORIZONTAL':
            uv_v = [size.y]
        elif path_type == 'VERTICAL':
            uv_v = [size.y]
        else:
            dx = size.x / (1 + self.subdiv_x)
            dy = size.y / (1 + self.subdiv_y)
            uv_v = []
            for i in range(self.subdiv_y + 1):
                uv_v.append(dy * (i + 1))
            for i in range(self.subdiv_x + 1):
                uv_v.append(dx * (i + 1))
            for i in range(self.subdiv_y + 1):
                uv_v.append(dy * (i + 1))
            for i in range(self.subdiv_x + 1):
                uv_v.append(dx * (i + 1))
            # uv_v = [size.y, size.x, size.y, size.x]

        uv_u = self.uv_u
        if self.closed_shape:
            n_pts = self.n_pts
        else:
            n_pts = self.n_pts - 1
        v0 = 0
        # uvs parties rondes
        for i in range(n_path_faces):
            v1 = v0 + uv_v[i]
            for j in range(n_pts):
                u0 = uv_u[j]
                u1 = uv_u[j + 1]
                uvs.append([(u0, v1), (u1, v1), (u1, v0), (u0, v0)])
            v0 = v1
        if self.side_cap_back > -1 or self.side_cap_front > -1:
            if path_type == 'ROUND':
                # rectangle with top part round
                coords = self._get_arc_coords(steps, Vector((0, 0, 0)), center,
                    origin, size, Vector((radius.x - x_cap, 0)), x_cap, pivot, x_cap)
            elif path_type == 'CIRCLE':
                # full circle
                coords = self._get_circle_coords(steps, Vector((0, 0, 0)), center,
                    origin, Vector((radius.x - x_cap, 0)))
            elif path_type == 'ELLIPSIS':
                coords = self._get_ellispe_coords(steps, Vector((0, 0, 0)), center,
                    origin, size, Vector((radius.x - x_cap, radius.y - x_cap)), x_cap, pivot, x_cap)
            elif path_type == 'QUADRI':
                coords = self._get_vertical_rectangular_trapezoid_coords(Vector((0, 0, 0)), center,
                    origin, size, radius.x, x_cap, pivot)
                # coords = self._get_trapezoidal_coords(0, origin, size, angle_y, x_cap, pivot, x_cap)
            else:
                coords = self._get_rectangular_coords(Vector((0, 0, 0)), size, x_cap, pivot, 0)
            if self.side_cap_front > -1:
                uvs.append(list(coords))
            if self.side_cap_back > -1:
                uvs.append(list(reversed(coords)))

        if self.closed_shape and not self.closed_path:
            coords = [(self.x[self.index[i]], y) for i, y in enumerate(self.y)]
            uvs.append(coords)
            uvs.append(list(reversed(coords)))
        return uvs

    ############################
    # Material indexes
    ############################

    def mat(self, steps, cap_front_id, cap_back_id, path_type='ROUND'):
        n_path_verts, n_path_faces = self.path_sections(steps, path_type)
        n_profil_faces = self.profil_faces
        idmat = []
        for i in range(n_path_faces):
            for f in range(n_profil_faces):
                idmat.append(self.idmat[f])
        if self.side_cap_front > -1:
            idmat.append(cap_front_id)
        if self.side_cap_back > -1:
            idmat.append(cap_back_id)
        if self.closed_shape and not self.closed_path:
            idmat.append(self.idmat[0])
            idmat.append(self.idmat[0])
        return idmat



#!/usr/bin/env python3
# Apache License, Version 2.0

import argparse
import os
import subprocess
import sys


def screenshot():
    import bpy

    output_path = sys.argv[-1]

    # Force redraw and take screenshot.
    bpy.ops.wm.redraw_timer(type='DRAW_WIN_SWAP', iterations=1)
    bpy.ops.screen.screenshot(filepath=output_path, full=True)

    bpy.ops.wm.quit_blender()


# When run from inside Blender, take screenshot and exit.
try:
    import bpy
    inside_blender = True
except ImportError:
    inside_blender = False

if inside_blender:
    screenshot()
    sys.exit(0)


def get_arguments(filepath, output_filepath):
    return [
        "--no-window-focus",
        "--window-geometry",
        "0", "0", "1024", "768",
        "-noaudio",
        "--factory-startup",
        "--enable-autoexec",
        "--debug-memory",
        "--debug-exit-on-error",
        filepath,
        "-P",
        os.path.realpath(__file__),
        "--",
        output_filepath + '0001.png']


def create_argparse():
    parser = argparse.ArgumentParser()
    parser.add_argument("-blender", nargs="+")
    parser.add_argument("-testdir", nargs=1)
    parser.add_argument("-outdir", nargs=1)
    parser.add_argument("-idiff", nargs=1)
    return parser


def main():
    parser = create_argparse()
    args = parser.parse_args()

    blender = args.blender[0]
    test_dir = args.testdir[0]
    idiff = args.idiff[0]
    output_dir = args.outdir[0]

    from modules import render_report
    report = render_report.Report("OpenGL Draw", output_dir, idiff)
    ok = report.run(test_dir, blender, get_arguments)

    sys.exit(not ok)


if __name__ == "__main__":
    main()



import gpu
import bgl

from gpu_extras.batch import batch_for_shader

class BL_UI_Widget:

    def __init__(self, x, y, width, height):
        self.x = x
        self.y = y
        self.x_screen = x
        self.y_screen = y
        self.width = width
        self.height = height
        self._bg_color = (0.8, 0.8, 0.8, 1.0)
        self._tag = None
        self.context = None
        self.__inrect = False
        self._mouse_down = False
        self._mouse_down_right = False
        self._is_visible = True

    def set_location(self, x, y):
        self.x = x
        self.y = y
        self.x_screen = x
        self.y_screen = y
        self.update(x,y)

    @property
    def bg_color(self):
        return self._bg_color

    @bg_color.setter
    def bg_color(self, value):
        self._bg_color = value

    @property
    def visible(self):
        return self._is_visible

    @visible.setter
    def visible(self, value):
        self._is_visible = value

    @property
    def tag(self):
        return self._tag

    @tag.setter
    def tag(self, value):
        self._tag = value

    def draw(self):
        if not self.visible:
            return

        self.shader.bind()
        self.shader.uniform_float("color", self._bg_color)

        bgl.glEnable(bgl.GL_BLEND)
        self.batch_panel.draw(self.shader)
        bgl.glDisable(bgl.GL_BLEND)

    def init(self, context):
        self.context = context
        self.update(self.x, self.y)

    def update(self, x, y):

        area_height = self.get_area_height()

        self.x_screen = x
        self.y_screen = y

        indices = ((0, 1, 2), (0, 2, 3))

        y_screen_flip = area_height - self.y_screen

        # bottom left, top left, top right, bottom right
        vertices = (
                    (self.x_screen, y_screen_flip),
                    (self.x_screen, y_screen_flip - self.height),
                    (self.x_screen + self.width, y_screen_flip - self.height),
                    (self.x_screen + self.width, y_screen_flip))

        self.shader = gpu.shader.from_builtin('2D_UNIFORM_COLOR')
        self.batch_panel = batch_for_shader(self.shader, 'TRIS', {"pos" : vertices}, indices=indices)

    def handle_event(self, event):
        x = event.mouse_region_x
        y = event.mouse_region_y

        if (event.type == 'LEFTMOUSE'):
            if (event.value == 'PRESS'):
                self._mouse_down = True
                return self.mouse_down(x, y)
            else:
                self._mouse_down = False
                self.mouse_up(x, y)

        elif (event.type == 'RIGHTMOUSE'):
            if (event.value == 'PRESS'):
                self._mouse_down_right = True
                return self.mouse_down_right(x, y)
            else:
                self._mouse_down_right = False
                self.mouse_up(x, y)

        elif (event.type == 'MOUSEMOVE'):
            self.mouse_move(x, y)

            inrect = self.is_in_rect(x, y)

            # we enter the rect
            if not self.__inrect and inrect:
                self.__inrect = True
                self.mouse_enter(event, x, y)

            # we are leaving the rect
            elif self.__inrect and not inrect:
                self.__inrect = False
                self.mouse_exit(event, x, y)

            return False

        elif event.value == 'PRESS' and (event.ascii != '' or event.type in self.get_input_keys()):
            return self.text_input(event)

        return False

    def get_input_keys(self)                :
        return []

    def get_area_height(self):
        return self.context.area.height

    def is_in_rect(self, x, y):
        area_height = self.get_area_height()

        widget_y = area_height - self.y_screen
        if (
            (self.x_screen <= x <= (self.x_screen + self.width)) and
            (widget_y >= y >= (widget_y - self.height))
            ):
            return True

        return False

    def text_input(self, event):
        return False

    def mouse_down(self, x, y):
        return self.is_in_rect(x,y)

    def mouse_down_right(self, x, y):
        return self.is_in_rect(x,y)

    def mouse_up(self, x, y):
        pass

    def set_mouse_enter(self, mouse_enter_func):
        self.mouse_enter_func = mouse_enter_func

    def call_mouse_enter(self):
        try:
            if self.mouse_enter_func:
                self.mouse_enter_func(self)
        except:
            pass

    def mouse_enter(self, event, x, y):
        self.call_mouse_enter()

    def set_mouse_exit(self, mouse_exit_func):
        self.mouse_exit_func = mouse_exit_func

    def call_mouse_exit(self):
        try:
            if self.mouse_exit_func:
                self.mouse_exit_func(self)
        except:
            pass

    def mouse_exit(self, event, x, y):
        self.call_mouse_exit()

    def mouse_move(self, x, y):
        pass



#!/usr/bin/env python3

# ***** BEGIN GPL LICENSE BLOCK *****
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Contributor(s): Campbell Barton
#
# #**** END GPL LICENSE BLOCK #****

# <pep8 compliant>

"""
this script updates XML themes once new settings are added

  ./blender.bin --background -noaudio --python source/tools/check_source/check_descriptions.py
"""

import bpy

DUPLICATE_WHITELIST = (
    # operators
    ('ACTION_OT_clean', 'GRAPH_OT_clean'),
    ('ACTION_OT_clickselect', 'GRAPH_OT_clickselect'),
    ('ACTION_OT_copy', 'GRAPH_OT_copy'),
    ('ACTION_OT_delete', 'GRAPH_OT_delete'),
    ('ACTION_OT_duplicate', 'GRAPH_OT_duplicate'),
    ('ACTION_OT_duplicate_move', 'GRAPH_OT_duplicate_move'),
    ('ACTION_OT_extrapolation_type', 'GRAPH_OT_extrapolation_type'),
    ('ACTION_OT_handle_type', 'GRAPH_OT_handle_type'),
    ('ACTION_OT_interpolation_type', 'GRAPH_OT_interpolation_type'),
    ('ACTION_OT_keyframe_insert', 'GRAPH_OT_keyframe_insert'),
    ('ACTION_OT_mirror', 'GRAPH_OT_mirror'),
    ('ACTION_OT_paste', 'GRAPH_OT_paste'),
    ('ACTION_OT_sample', 'GRAPH_OT_sample'),
    ('ACTION_OT_select_all', 'GRAPH_OT_select_all'),
    ('ACTION_OT_select_border', 'GRAPH_OT_select_border'),
    ('ACTION_OT_select_column', 'GRAPH_OT_select_column'),
    ('ACTION_OT_select_leftright', 'GRAPH_OT_select_leftright'),
    ('ACTION_OT_select_less', 'GRAPH_OT_select_less'),
    ('ACTION_OT_select_linked', 'GRAPH_OT_select_linked'),
    ('ACTION_OT_select_more', 'GRAPH_OT_select_more'),
    ('ACTION_OT_unlink', 'NLA_OT_action_unlink'),
    ('ACTION_OT_view_all', 'CLIP_OT_dopesheet_view_all', 'GRAPH_OT_view_all'),
    ('ACTION_OT_view_frame', 'GRAPH_OT_view_frame'),
    ('ANIM_OT_change_frame', 'CLIP_OT_change_frame', 'IMAGE_OT_change_frame'),
    ('ARMATURE_OT_autoside_names', 'POSE_OT_autoside_names'),
    ('ARMATURE_OT_bone_layers', 'POSE_OT_bone_layers'),
    ('ARMATURE_OT_extrude_forked', 'ARMATURE_OT_extrude_move'),
    ('ARMATURE_OT_flip_names', 'POSE_OT_flip_names'),
    ('ARMATURE_OT_select_all', 'POSE_OT_select_all'),
    ('ARMATURE_OT_select_hierarchy', 'POSE_OT_select_hierarchy'),
    ('ARMATURE_OT_select_linked', 'POSE_OT_select_linked'),
    ('ARMATURE_OT_select_mirror', 'POSE_OT_select_mirror'),
    ('CLIP_OT_cursor_set', 'UV_OT_cursor_set'),
    ('CLIP_OT_disable_markers', 'CLIP_OT_graph_disable_markers'),
    ('CLIP_OT_graph_select_border', 'MASK_OT_select_border'),
    ('CLIP_OT_view_ndof', 'IMAGE_OT_view_ndof', 'VIEW2D_OT_ndof'),
    ('CLIP_OT_view_pan', 'IMAGE_OT_view_pan', 'VIEW2D_OT_pan', 'VIEW3D_OT_view_pan'),
    ('CLIP_OT_view_zoom', 'VIEW2D_OT_zoom'),
    ('CLIP_OT_view_zoom_in', 'VIEW2D_OT_zoom_in'),
    ('CLIP_OT_view_zoom_out', 'VIEW2D_OT_zoom_out'),
    ('CONSOLE_OT_copy', 'FONT_OT_text_copy', 'TEXT_OT_copy'),
    ('CONSOLE_OT_delete', 'FONT_OT_delete', 'TEXT_OT_delete'),
    ('CONSOLE_OT_insert', 'FONT_OT_text_insert', 'TEXT_OT_insert'),
    ('CONSOLE_OT_paste', 'FONT_OT_text_paste', 'TEXT_OT_paste'),
    ('CURVE_OT_handle_type_set', 'MASK_OT_handle_type_set'),
    ('CURVE_OT_shortest_path_pick', 'MESH_OT_shortest_path_pick'),
    ('CURVE_OT_switch_direction', 'MASK_OT_switch_direction'),
    ('FONT_OT_line_break', 'TEXT_OT_line_break'),
    ('FONT_OT_move', 'TEXT_OT_move'),
    ('FONT_OT_move_select', 'TEXT_OT_move_select'),
    ('FONT_OT_select_all', 'TEXT_OT_select_all'),
    ('FONT_OT_text_cut', 'TEXT_OT_cut'),
    ('GRAPH_OT_previewrange_set', 'NLA_OT_previewrange_set'),
    ('GRAPH_OT_properties', 'IMAGE_OT_properties', 'LOGIC_OT_properties', 'NLA_OT_properties'),
    ('IMAGE_OT_clear_render_border', 'VIEW3D_OT_clear_render_border'),
    ('IMAGE_OT_render_border', 'VIEW3D_OT_render_border'),
    ('IMAGE_OT_toolshelf', 'NODE_OT_toolbar', 'VIEW3D_OT_toolshelf'),
    ('LATTICE_OT_select_ungrouped', 'MESH_OT_select_ungrouped', 'PAINT_OT_vert_select_ungrouped'),
    ('MESH_OT_extrude_region_move', 'MESH_OT_extrude_region_shrink_fatten'),
    ('NODE_OT_add_node', 'NODE_OT_add_search'),
    ('NODE_OT_move_detach_links', 'NODE_OT_move_detach_links_release'),
    ('NODE_OT_properties', 'VIEW3D_OT_properties'),
    ('OBJECT_OT_bake', 'OBJECT_OT_bake_image'),
    ('OBJECT_OT_duplicate_move', 'OBJECT_OT_duplicate_move_linked'),
    ('WM_OT_context_cycle_enum', 'WM_OT_context_toggle', 'WM_OT_context_toggle_enum'),
    ('WM_OT_context_set_boolean', 'WM_OT_context_set_enum', 'WM_OT_context_set_float',
     'WM_OT_context_set_int', 'WM_OT_context_set_string', 'WM_OT_context_set_value'),
)

DUPLICATE_IGNORE = {
    "",
}


def check_duplicates():
    import rna_info

    DUPLICATE_IGNORE_FOUND = set()
    DUPLICATE_WHITELIST_FOUND = set()

    structs, funcs, ops, props = rna_info.BuildRNAInfo()

    # This is mainly useful for operators,
    # other types have too many false positives

    # for t in (structs, funcs, ops, props):
    for t in (ops, ):
        description_dict = {}
        print("")
        for k, v in t.items():
            if v.description not in DUPLICATE_IGNORE:
                id_str = ".".join([s if isinstance(s, str) else s.identifier for s in k if s])
                description_dict.setdefault(v.description, []).append(id_str)
            else:
                DUPLICATE_IGNORE_FOUND.add(v.description)
        # sort for easier viewing
        sort_ls = [(tuple(sorted(v)), k) for k, v in description_dict.items()]
        sort_ls.sort()

        for v, k in sort_ls:
            if len(v) > 1:
                if v not in DUPLICATE_WHITELIST:
                    print("found %d: %r, \"%s\"" % (len(v), v, k))
                    #print("%r," % (v,))
                else:
                    DUPLICATE_WHITELIST_FOUND.add(v)

    test = (DUPLICATE_IGNORE - DUPLICATE_IGNORE_FOUND)
    if test:
        print("Invalid 'DUPLICATE_IGNORE': %r" % test)
    test = (set(DUPLICATE_WHITELIST) - DUPLICATE_WHITELIST_FOUND)
    if test:
        print("Invalid 'DUPLICATE_WHITELIST': %r" % test)


def main():
    check_duplicates()


if __name__ == "__main__":
    main()



"""
Persistent Handler Example
++++++++++++++++++++++++++

By default handlers are freed when loading new files, in some cases you may
want the handler stay running across multiple files (when the handler is
part of an add-on for example).

For this the :data:`bpy.app.handlers.persistent` decorator needs to be used.
"""

import bpy
from bpy.app.handlers import persistent


@persistent
def load_handler(dummy):
    print("Load Handler:", bpy.data.filepath)


bpy.app.handlers.load_post.append(load_handler)



# -*- coding:utf-8 -*-

# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110- 1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# <pep8 compliant>

# ----------------------------------------------------------
# Author: Stephen Leger (s-leger)
#
# ----------------------------------------------------------
# noinspection PyUnresolvedReferences
import bpy
# noinspection PyUnresolvedReferences
from bpy.types import Operator, PropertyGroup, Mesh, Panel
from bpy.props import (
    FloatProperty, BoolProperty, IntProperty, CollectionProperty,
    StringProperty, EnumProperty, FloatVectorProperty
    )
from .bmesh_utils import BmeshEdit as bmed
from .panel import Panel as Lofter
from mathutils import Vector, Matrix
from mathutils.geometry import interpolate_bezier
from math import sin, cos, pi, acos, atan2
from .archipack_manipulator import Manipulable, archipack_manipulator
from .archipack_2d import Line, Arc
from .archipack_preset import ArchipackPreset, PresetMenuOperator
from .archipack_object import ArchipackCreateTool, ArchipackObject


class Fence():

    def __init__(self):
        # total distance from start
        self.dist = 0
        self.t_start = 0
        self.t_end = 0
        self.dz = 0
        self.z0 = 0

    def set_offset(self, offset, last=None):
        """
            Offset line and compute intersection point
            between segments
        """
        self.line = self.make_offset(offset, last)

    @property
    def t_diff(self):
        return self.t_end - self.t_start

    def straight_fence(self, a0, length):
        s = self.straight(length).rotate(a0)
        return StraightFence(s.p, s.v)

    def curved_fence(self, a0, da, radius):
        n = self.normal(1).rotate(a0).scale(radius)
        if da < 0:
            n.v = -n.v
        a0 = n.angle
        c = n.p - n.v
        return CurvedFence(c, radius, a0, da)


class StraightFence(Fence, Line):
    def __str__(self):
        return "t_start:{} t_end:{} dist:{}".format(self.t_start, self.t_end, self.dist)

    def __init__(self, p, v):
        Fence.__init__(self)
        Line.__init__(self, p, v)


class CurvedFence(Fence, Arc):
    def __str__(self):
        return "t_start:{} t_end:{} dist:{}".format(self.t_start, self.t_end, self.dist)

    def __init__(self, c, radius, a0, da):
        Fence.__init__(self)
        Arc.__init__(self, c, radius, a0, da)


class FenceSegment():
    def __str__(self):
        return "t_start:{} t_end:{} n_step:{}  t_step:{} i_start:{} i_end:{}".format(
            self.t_start, self.t_end, self.n_step, self.t_step, self.i_start, self.i_end)

    def __init__(self, t_start, t_end, n_step, t_step, i_start, i_end):
        self.t_start = t_start
        self.t_end = t_end
        self.n_step = n_step
        self.t_step = t_step
        self.i_start = i_start
        self.i_end = i_end


class FenceGenerator():

    def __init__(self, parts):
        self.parts = parts
        self.segs = []
        self.length = 0
        self.user_defined_post = None
        self.user_defined_uvs = None
        self.user_defined_mat = None

    def add_part(self, part):

        if len(self.segs) < 1:
            s = None
        else:
            s = self.segs[-1]

        # start a new fence
        if s is None:
            if part.type == 'S_FENCE':
                p = Vector((0, 0))
                v = part.length * Vector((cos(part.a0), sin(part.a0)))
                s = StraightFence(p, v)
            elif part.type == 'C_FENCE':
                c = -part.radius * Vector((cos(part.a0), sin(part.a0)))
                s = CurvedFence(c, part.radius, part.a0, part.da)
        else:
            if part.type == 'S_FENCE':
                s = s.straight_fence(part.a0, part.length)
            elif part.type == 'C_FENCE':
                s = s.curved_fence(part.a0, part.da, part.radius)

        # s.dist = self.length
        # self.length += s.length
        self.segs.append(s)
        self.last_type = type

    def set_offset(self, offset):
        # @TODO:
        # re-evaluate length of offset line here
        last = None
        for seg in self.segs:
            seg.set_offset(offset, last)
            last = seg.line

    def param_t(self, angle_limit, post_spacing):
        """
            setup corners and fences dz
            compute index of fences which belong to each group of fences between corners
            compute t of each fence
        """
        # segments are group of parts separated by limit angle
        self.segments = []
        i_start = 0
        t_start = 0
        dist_0 = 0
        z = 0
        self.length = 0
        n_parts = len(self.parts) - 1
        for i, f in enumerate(self.segs):
            f.dist = self.length
            self.length += f.line.length

        vz0 = Vector((1, 0))
        angle_z = 0
        for i, f in enumerate(self.segs):
            dz = self.parts[i].dz
            if f.dist > 0:
                f.t_start = f.dist / self.length
            else:
                f.t_start = 0

            f.t_end = (f.dist + f.line.length) / self.length
            f.z0 = z
            f.dz = dz
            z += dz

            if i < n_parts:

                vz1 = Vector((self.segs[i + 1].length, self.parts[i + 1].dz))
                angle_z = abs(vz0.angle_signed(vz1))
                vz0 = vz1

                if (abs(self.parts[i + 1].a0) >= angle_limit or angle_z >= angle_limit):
                    l_seg = f.dist + f.line.length - dist_0
                    t_seg = f.t_end - t_start
                    n_fences = max(1, int(l_seg / post_spacing))
                    t_fence = t_seg / n_fences
                    segment = FenceSegment(t_start, f.t_end, n_fences, t_fence, i_start, i)
                    dist_0 = f.dist + f.line.length
                    t_start = f.t_end
                    i_start = i
                    self.segments.append(segment)

            manipulators = self.parts[i].manipulators
            p0 = f.line.p0.to_3d()
            p1 = f.line.p1.to_3d()
            # angle from last to current segment
            if i > 0:
                v0 = self.segs[i - 1].line.straight(-1, 1).v.to_3d()
                v1 = f.line.straight(1, 0).v.to_3d()
                manipulators[0].set_pts([p0, v0, v1])

            if type(f).__name__ == "StraightFence":
                # segment length
                manipulators[1].type_key = 'SIZE'
                manipulators[1].prop1_name = "length"
                manipulators[1].set_pts([p0, p1, (1, 0, 0)])
            else:
                # segment radius + angle
                v0 = (f.line.p0 - f.c).to_3d()
                v1 = (f.line.p1 - f.c).to_3d()
                manipulators[1].type_key = 'ARC_ANGLE_RADIUS'
                manipulators[1].prop1_name = "da"
                manipulators[1].prop2_name = "radius"
                manipulators[1].set_pts([f.c.to_3d(), v0, v1])

            # snap manipulator, don't change index !
            manipulators[2].set_pts([p0, p1, (1, 0, 0)])

        f = self.segs[-1]
        l_seg = f.dist + f.line.length - dist_0
        t_seg = f.t_end - t_start
        n_fences = max(1, int(l_seg / post_spacing))
        t_fence = t_seg / n_fences
        segment = FenceSegment(t_start, f.t_end, n_fences, t_fence, i_start, len(self.segs) - 1)
        self.segments.append(segment)

    def setup_user_defined_post(self, o, post_x, post_y, post_z):
        self.user_defined_post = o
        x = o.bound_box[6][0] - o.bound_box[0][0]
        y = o.bound_box[6][1] - o.bound_box[0][1]
        z = o.bound_box[6][2] - o.bound_box[0][2]
        self.user_defined_post_scale = Vector((post_x / x, post_y / -y, post_z / z))
        m = o.data
        # create vertex group lookup dictionary for names
        vgroup_names = {vgroup.index: vgroup.name for vgroup in o.vertex_groups}
        # create dictionary of vertex group assignments per vertex
        self.vertex_groups = [[vgroup_names[g.group] for g in v.groups] for v in m.vertices]
        # uvs
        uv_act = m.uv_layers.active
        if uv_act is not None:
            uv_layer = uv_act.data
            self.user_defined_uvs = [[uv_layer[li].uv for li in p.loop_indices] for p in m.polygons]
        else:
            self.user_defined_uvs = [[(0, 0) for i in p.vertices] for p in m.polygons]
        # material ids
        self.user_defined_mat = [p.material_index for p in m.polygons]

    def get_user_defined_post(self, tM, z0, z1, z2, slope, post_z, verts, faces, matids, uvs):
        f = len(verts)
        m = self.user_defined_post.data
        for i, g in enumerate(self.vertex_groups):
            co = m.vertices[i].co.copy()
            co.x *= self.user_defined_post_scale.x
            co.y *= self.user_defined_post_scale.y
            co.z *= self.user_defined_post_scale.z
            if 'Slope' in g:
                co.z += co.y * slope
            verts.append(tM @ co)
        matids += self.user_defined_mat
        faces += [tuple([i + f for i in p.vertices]) for p in m.polygons]
        uvs += self.user_defined_uvs

    def get_post(self, post, post_x, post_y, post_z, post_alt, sub_offset_x,
            id_mat, verts, faces, matids, uvs):

        n, dz, zl = post
        slope = dz * post_y

        if self.user_defined_post is not None:
            x, y = -n.v.normalized()
            p = n.p + sub_offset_x * n.v.normalized()
            tM = Matrix([
                [x, y, 0, p.x],
                [y, -x, 0, p.y],
                [0, 0, 1, zl + post_alt],
                [0, 0, 0, 1]
            ])
            self.get_user_defined_post(tM, zl, 0, 0, dz, post_z, verts, faces, matids, uvs)
            return

        z3 = zl + post_z + post_alt - slope
        z4 = zl + post_z + post_alt + slope
        z0 = zl + post_alt - slope
        z1 = zl + post_alt + slope
        vn = n.v.normalized()
        dx = post_x * vn
        dy = post_y * Vector((vn.y, -vn.x))
        oy = sub_offset_x * vn
        x0, y0 = n.p - dx + dy + oy
        x1, y1 = n.p - dx - dy + oy
        x2, y2 = n.p + dx - dy + oy
        x3, y3 = n.p + dx + dy + oy
        f = len(verts)
        verts.extend([(x0, y0, z0), (x0, y0, z3),
                    (x1, y1, z1), (x1, y1, z4),
                    (x2, y2, z1), (x2, y2, z4),
                    (x3, y3, z0), (x3, y3, z3)])
        faces.extend([(f, f + 1, f + 3, f + 2),
                    (f + 2, f + 3, f + 5, f + 4),
                    (f + 4, f + 5, f + 7, f + 6),
                    (f + 6, f + 7, f + 1, f),
                    (f, f + 2, f + 4, f + 6),
                    (f + 7, f + 5, f + 3, f + 1)])
        matids.extend([id_mat, id_mat, id_mat, id_mat, id_mat, id_mat])
        x = [(0, 0), (0, post_z), (post_x, post_z), (post_x, 0)]
        y = [(0, 0), (0, post_z), (post_y, post_z), (post_y, 0)]
        z = [(0, 0), (post_x, 0), (post_x, post_y), (0, post_y)]
        uvs.extend([x, y, x, y, z, z])

    def get_panel(self, subs, altitude, panel_x, panel_z, sub_offset_x, idmat, verts, faces, matids, uvs):
        n_subs = len(subs)
        if n_subs < 1:
            return
        f = len(verts)
        x0 = sub_offset_x - 0.5 * panel_x
        x1 = sub_offset_x + 0.5 * panel_x
        z0 = 0
        z1 = panel_z
        profile = [Vector((x0, z0)), Vector((x1, z0)), Vector((x1, z1)), Vector((x0, z1))]
        user_path_uv_v = []
        n_sections = n_subs - 1
        n, dz, zl = subs[0]
        p0 = n.p
        v0 = n.v.normalized()
        for s, section in enumerate(subs):
            n, dz, zl = section
            p1 = n.p
            if s < n_sections:
                v1 = subs[s + 1][0].v.normalized()
            dir = (v0 + v1).normalized()
            scale = 1 / cos(0.5 * acos(min(1, max(-1, v0.dot(v1)))))
            for p in profile:
                x, y = n.p + scale * p.x * dir
                z = zl + p.y + altitude
                verts.append((x, y, z))
            if s > 0:
                user_path_uv_v.append((p1 - p0).length)
            p0 = p1
            v0 = v1

        # build faces using Panel
        lofter = Lofter(
            # closed_shape, index, x, y, idmat
            True,
            [i for i in range(len(profile))],
            [p.x for p in profile],
            [p.y for p in profile],
            [idmat for i in range(len(profile))],
            closed_path=False,
            user_path_uv_v=user_path_uv_v,
            user_path_verts=n_subs
            )
        faces += lofter.faces(16, offset=f, path_type='USER_DEFINED')
        matids += lofter.mat(16, idmat, idmat, path_type='USER_DEFINED')
        v = Vector((0, 0))
        uvs += lofter.uv(16, v, v, v, v, 0, v, 0, 0, path_type='USER_DEFINED')

    def make_subs(self, x, y, z, post_y, altitude,
            sub_spacing, offset_x, sub_offset_x, mat, verts, faces, matids, uvs):

        t_post = (0.5 * post_y - y) / self.length
        t_spacing = (sub_spacing + y) / self.length

        for segment in self.segments:
            t_step = segment.t_step
            t_start = segment.t_start + t_post
            s = 0
            s_sub = t_step - 2 * t_post
            n_sub = int(s_sub / t_spacing)
            if n_sub > 0:
                t_sub = s_sub / n_sub
            else:
                t_sub = 1
            i = segment.i_start
            while s < segment.n_step:
                t_cur = t_start + s * t_step
                for j in range(1, n_sub):
                    t_s = t_cur + t_sub * j
                    while self.segs[i].t_end < t_s:
                        i += 1
                    f = self.segs[i]
                    t = (t_s - f.t_start) / f.t_diff
                    n = f.line.normal(t)
                    post = (n, f.dz / f.length, f.z0 + f.dz * t)
                    self.get_post(post, x, y, z, altitude, sub_offset_x, mat, verts, faces, matids, uvs)
                s += 1

    def make_post(self, x, y, z, altitude, x_offset, mat, verts, faces, matids, uvs):

        for segment in self.segments:
            t_step = segment.t_step
            t_start = segment.t_start
            s = 0
            i = segment.i_start
            while s < segment.n_step:
                t_cur = t_start + s * t_step
                while self.segs[i].t_end < t_cur:
                    i += 1
                f = self.segs[i]
                t = (t_cur - f.t_start) / f.t_diff
                n = f.line.normal(t)
                post = (n, f.dz / f.line.length, f.z0 + f.dz * t)
                # self.get_post(post, x, y, z, altitude, x_offset, mat, verts, faces, matids, uvs)
                self.get_post(post, x, y, z, altitude, 0, mat, verts, faces, matids, uvs)
                s += 1

            if segment.i_end + 1 == len(self.segs):
                f = self.segs[segment.i_end]
                n = f.line.normal(1)
                post = (n, f.dz / f.line.length, f.z0 + f.dz)
                # self.get_post(post, x, y, z, altitude, x_offset, mat, verts, faces, matids, uvs)
                self.get_post(post, x, y, z, altitude, 0, mat, verts, faces, matids, uvs)

    def make_panels(self, x, z, post_y, altitude, panel_dist,
            offset_x, sub_offset_x, idmat, verts, faces, matids, uvs):

        t_post = (0.5 * post_y + panel_dist) / self.length
        for segment in self.segments:
            t_step = segment.t_step
            t_start = segment.t_start
            s = 0
            i = segment.i_start
            while s < segment.n_step:
                subs = []
                t_cur = t_start + s * t_step + t_post
                t_end = t_start + (s + 1) * t_step - t_post
                # find first section
                while self.segs[i].t_end < t_cur and i < segment.i_end:
                    i += 1
                f = self.segs[i]
                # 1st section
                t = (t_cur - f.t_start) / f.t_diff
                n = f.line.normal(t)
                subs.append((n, f.dz / f.line.length, f.z0 + f.dz * t))
                # crossing sections -> new segment
                while i < segment.i_end:
                    f = self.segs[i]
                    if f.t_end < t_end:
                        if type(f).__name__ == 'CurvedFence':
                            # can't end after segment
                            t0 = max(0, (t_cur - f.t_start) / f.t_diff)
                            t1 = min(1, (t_end - f.t_start) / f.t_diff)
                            n_s = int(max(1, abs(f.da) * (5) / pi - 1))
                            dt = (t1 - t0) / n_s
                            for j in range(1, n_s + 1):
                                t = t0 + dt * j
                                n = f.line.sized_normal(t, 1)
                                # n.p = f.lerp(x_offset)
                                subs.append((n, f.dz / f.line.length, f.z0 + f.dz * t))
                        else:
                            n = f.line.normal(1)
                            subs.append((n, f.dz / f.line.length, f.z0 + f.dz))
                    if f.t_end >= t_end:
                        break
                    elif f.t_start < t_end:
                        i += 1

                f = self.segs[i]
                # last section
                if type(f).__name__ == 'CurvedFence':
                    # can't start before segment
                    t0 = max(0, (t_cur - f.t_start) / f.t_diff)
                    t1 = min(1, (t_end - f.t_start) / f.t_diff)
                    n_s = int(max(1, abs(f.da) * (5) / pi - 1))
                    dt = (t1 - t0) / n_s
                    for j in range(1, n_s + 1):
                        t = t0 + dt * j
                        n = f.line.sized_normal(t, 1)
                        # n.p = f.lerp(x_offset)
                        subs.append((n, f.dz / f.line.length, f.z0 + f.dz * t))
                else:
                    t = (t_end - f.t_start) / f.t_diff
                    n = f.line.normal(t)
                    subs.append((n, f.dz / f.line.length, f.z0 + f.dz * t))

                # self.get_panel(subs, altitude, x, z, 0, idmat, verts, faces, matids, uvs)
                self.get_panel(subs, altitude, x, z, sub_offset_x, idmat, verts, faces, matids, uvs)
                s += 1

    def make_profile(self, profile, idmat,
            x_offset, z_offset, extend, verts, faces, matids, uvs):

        last = None
        for seg in self.segs:
            seg.p_line = seg.make_offset(x_offset, last)
            last = seg.p_line

        n_fences = len(self.segs) - 1

        if n_fences < 0:
            return

        sections = []

        f = self.segs[0]

        # first step
        if extend != 0 and f.p_line.length != 0:
            t = -extend / self.segs[0].p_line.length
            n = f.p_line.sized_normal(t, 1)
            # n.p = f.lerp(x_offset)
            sections.append((n, f.dz / f.p_line.length, f.z0 + f.dz * t))

        # add first section
        n = f.p_line.sized_normal(0, 1)
        # n.p = f.lerp(x_offset)
        sections.append((n, f.dz / f.p_line.length, f.z0))

        for s, f in enumerate(self.segs):
            if f.p_line.length == 0:
                continue
            if type(f).__name__ == 'CurvedFence':
                n_s = int(max(1, abs(f.da) * 30 / pi - 1))
                for i in range(1, n_s + 1):
                    t = i / n_s
                    n = f.p_line.sized_normal(t, 1)
                    # n.p = f.lerp(x_offset)
                    sections.append((n, f.dz / f.p_line.length, f.z0 + f.dz * t))
            else:
                n = f.p_line.sized_normal(1, 1)
                # n.p = f.lerp(x_offset)
                sections.append((n, f.dz / f.p_line.length, f.z0 + f.dz))

        if extend != 0 and f.p_line.length != 0:
            t = 1 + extend / self.segs[-1].p_line.length
            n = f.p_line.sized_normal(t, 1)
            # n.p = f.lerp(x_offset)
            sections.append((n, f.dz / f.p_line.length, f.z0 + f.dz * t))

        user_path_verts = len(sections)
        offset = len(verts)
        if user_path_verts > 0:
            user_path_uv_v = []
            n, dz, z0 = sections[-1]
            sections[-1] = (n, dz, z0)
            n_sections = user_path_verts - 1
            n, dz, zl = sections[0]
            p0 = n.p
            v0 = n.v.normalized()
            for s, section in enumerate(sections):
                n, dz, zl = section
                p1 = n.p
                if s < n_sections:
                    v1 = sections[s + 1][0].v.normalized()
                dir = (v0 + v1).normalized()
                scale = min(10, 1 / cos(0.5 * acos(min(1, max(-1, v0.dot(v1) )))))
                for p in profile:
                    # x, y = n.p + scale * (x_offset + p.x) * dir
                    x, y = n.p + scale * p.x * dir
                    z = zl + p.y + z_offset
                    verts.append((x, y, z))
                if s > 0:
                    user_path_uv_v.append((p1 - p0).length)
                p0 = p1
                v0 = v1

            # build faces using Panel
            lofter = Lofter(
                # closed_shape, index, x, y, idmat
                True,
                [i for i in range(len(profile))],
                [p.x for p in profile],
                [p.y for p in profile],
                [idmat for i in range(len(profile))],
                closed_path=False,
                user_path_uv_v=user_path_uv_v,
                user_path_verts=user_path_verts
                )
            faces += lofter.faces(16, offset=offset, path_type='USER_DEFINED')
            matids += lofter.mat(16, idmat, idmat, path_type='USER_DEFINED')
            v = Vector((0, 0))
            uvs += lofter.uv(16, v, v, v, v, 0, v, 0, 0, path_type='USER_DEFINED')


def update(self, context):
    self.update(context)


def update_manipulators(self, context):
    self.update(context, manipulable_refresh=True)


def update_path(self, context):
    self.update_path(context)


def update_type(self, context):

    d = self.find_datablock_in_selection(context)

    if d is not None and d.auto_update:

        d.auto_update = False
        # find part index
        idx = 0
        for i, part in enumerate(d.parts):
            if part == self:
                idx = i
                break
        part = d.parts[idx]
        a0 = 0
        if idx > 0:
            g = d.get_generator()
            w0 = g.segs[idx - 1]
            a0 = w0.straight(1).angle
            if "C_" in self.type:
                w = w0.straight_fence(part.a0, part.length)
            else:
                w = w0.curved_fence(part.a0, part.da, part.radius)
        else:
            if "C_" in self.type:
                p = Vector((0, 0))
                v = self.length * Vector((cos(self.a0), sin(self.a0)))
                w = StraightFence(p, v)
                a0 = pi / 2
            else:
                c = -self.radius * Vector((cos(self.a0), sin(self.a0)))
                w = CurvedFence(c, self.radius, self.a0, pi)

        # not closed, see wall
        # for closed ability
        dp = w.p1 - w.p0

        if "C_" in self.type:
            part.radius = 0.5 * dp.length
            part.da = pi
            a0 = atan2(dp.y, dp.x) - pi / 2 - a0
        else:
            part.length = dp.length
            a0 = atan2(dp.y, dp.x) - a0

        if a0 > pi:
            a0 -= 2 * pi
        if a0 < -pi:
            a0 += 2 * pi
        part.a0 = a0

        if idx + 1 < d.n_parts:
            # adjust rotation of next part
            part1 = d.parts[idx + 1]
            if "C_" in part.type:
                a0 = part1.a0 - pi / 2
            else:
                a0 = part1.a0 + w.straight(1).angle - atan2(dp.y, dp.x)

            if a0 > pi:
                a0 -= 2 * pi
            if a0 < -pi:
                a0 += 2 * pi
            part1.a0 = a0

        d.auto_update = True


materials_enum = (
            ('0', 'Wood', '', 0),
            ('1', 'Metal', '', 1),
            ('2', 'Glass', '', 2)
            )


class archipack_fence_material(PropertyGroup):
    index : EnumProperty(
        items=materials_enum,
        default='0',
        update=update
        )

    def find_datablock_in_selection(self, context):
        """
            find witch selected object this instance belongs to
            provide support for "copy to selected"
        """
        selected = context.selected_objects[:]
        for o in selected:
            props = archipack_fence.datablock(o)
            if props:
                for part in props.rail_mat:
                    if part == self:
                        return props
        return None

    def update(self, context):
        props = self.find_datablock_in_selection(context)
        if props is not None:
            props.update(context)


class archipack_fence_part(PropertyGroup):
    type : EnumProperty(
            items=(
                ('S_FENCE', 'Straight fence', '', 0),
                ('C_FENCE', 'Curved fence', '', 1),
                ),
            default='S_FENCE',
            update=update_type
            )
    length : FloatProperty(
            name="Length",
            min=0.01,
            default=2.0,
            unit='LENGTH', subtype='DISTANCE',
            update=update
            )
    radius : FloatProperty(
            name="Radius",
            min=0.01,
            default=0.7,
            unit='LENGTH', subtype='DISTANCE',
            update=update
            )
    da : FloatProperty(
            name="Angle",
            min=-pi,
            max=pi,
            default=pi / 2,
            subtype='ANGLE', unit='ROTATION',
            update=update
            )
    a0 : FloatProperty(
            name="Start angle",
            min=-2 * pi,
            max=2 * pi,
            default=0,
            subtype='ANGLE', unit='ROTATION',
            update=update
            )
    dz : FloatProperty(
            name="delta z",
            default=0,
            unit='LENGTH', subtype='DISTANCE'
            )

    manipulators : CollectionProperty(type=archipack_manipulator)

    def find_datablock_in_selection(self, context):
        """
            find witch selected object this instance belongs to
            provide support for "copy to selected"
        """
        selected = context.selected_objects[:]
        for o in selected:
            props = archipack_fence.datablock(o)
            if props is not None:
                for part in props.parts:
                    if part == self:
                        return props
        return None

    def update(self, context, manipulable_refresh=False):
        props = self.find_datablock_in_selection(context)
        if props is not None:
            props.update(context, manipulable_refresh)

    def draw(self, layout, context, index):
        box = layout.box()
        row = box.row()
        row.prop(self, "type", text=str(index + 1))
        if self.type in ['C_FENCE']:
            row = box.row()
            row.prop(self, "radius")
            row = box.row()
            row.prop(self, "da")
        else:
            row = box.row()
            row.prop(self, "length")
        row = box.row()
        row.prop(self, "a0")


class archipack_fence(ArchipackObject, Manipulable, PropertyGroup):

    parts : CollectionProperty(type=archipack_fence_part)
    user_defined_path : StringProperty(
            name="User defined",
            update=update_path
            )
    user_defined_spline : IntProperty(
            name="Spline index",
            min=0,
            default=0,
            update=update_path
            )
    user_defined_resolution : IntProperty(
            name="Resolution",
            min=1,
            max=128,
            default=12, update=update_path
            )
    n_parts : IntProperty(
            name="Parts",
            min=1,
            default=1, update=update_manipulators
            )
    x_offset : FloatProperty(
            name="Offset",
            default=0.0, precision=2, step=1,
            unit='LENGTH', subtype='DISTANCE',
            update=update
            )

    radius : FloatProperty(
            name="Radius",
            min=0.01,
            default=0.7,
            unit='LENGTH', subtype='DISTANCE',
            update=update
            )
    da : FloatProperty(
            name="Angle",
            min=-pi,
            max=pi,
            default=pi / 2,
            subtype='ANGLE', unit='ROTATION',
            update=update
            )
    angle_limit : FloatProperty(
            name="Angle",
            min=0,
            max=2 * pi,
            default=pi / 8,
            subtype='ANGLE', unit='ROTATION',
            update=update_manipulators
            )
    shape : EnumProperty(
            items=(
                ('RECTANGLE', 'Straight', '', 0),
                ('CIRCLE', 'Curved ', '', 1)
                ),
            default='RECTANGLE',
            update=update
            )
    post : BoolProperty(
            name='Enable',
            default=True,
            update=update
            )
    post_spacing : FloatProperty(
            name="Spacing",
            min=0.1,
            default=1.0, precision=2, step=1,
            unit='LENGTH', subtype='DISTANCE',
            update=update
            )
    post_x : FloatProperty(
            name="Width",
            min=0.001,
            default=0.04, precision=2, step=1,
            unit='LENGTH', subtype='DISTANCE',
            update=update
            )
    post_y : FloatProperty(
            name="Length",
            min=0.001, max=1000,
            default=0.04, precision=2, step=1,
            unit='LENGTH', subtype='DISTANCE',
            update=update
            )
    post_z : FloatProperty(
            name="Height",
            min=0.001,
            default=1, precision=2, step=1,
            unit='LENGTH', subtype='DISTANCE',
            update=update
            )
    post_alt : FloatProperty(
            name="Altitude",
            default=0, precision=2, step=1,
            unit='LENGTH', subtype='DISTANCE',
            update=update
            )
    user_defined_post_enable : BoolProperty(
            name="User",
            update=update,
            default=True
            )
    user_defined_post : StringProperty(
            name="User defined",
            update=update
            )
    idmat_post : EnumProperty(
            name="Post",
            items=materials_enum,
            default='1',
            update=update
            )
    subs : BoolProperty(
            name='Enable',
            default=False,
            update=update
            )
    subs_spacing : FloatProperty(
            name="Spacing",
            min=0.05,
            default=0.10, precision=2, step=1,
            unit='LENGTH', subtype='DISTANCE',
            update=update
            )
    subs_x : FloatProperty(
            name="Width",
            min=0.001,
            default=0.02, precision=2, step=1,
            unit='LENGTH', subtype='DISTANCE',
            update=update
            )
    subs_y : FloatProperty(
            name="Length",
            min=0.001,
            default=0.02, precision=2, step=1,
            unit='LENGTH', subtype='DISTANCE',
            update=update
            )
    subs_z : FloatProperty(
            name="Height",
            min=0.001,
            default=1, precision=2, step=1,
            unit='LENGTH', subtype='DISTANCE',
            update=update
            )
    subs_alt : FloatProperty(
            name="Altitude",
            default=0, precision=2, step=1,
            unit='LENGTH', subtype='DISTANCE',
            update=update
            )
    subs_offset_x : FloatProperty(
            name="Offset",
            default=0.0, precision=2, step=1,
            unit='LENGTH', subtype='DISTANCE',
            update=update
            )
    subs_bottom : EnumProperty(
            name="Bottom",
            items=(
                ('STEP', 'Follow step', '', 0),
                ('LINEAR', 'Linear', '', 1),
                ),
            default='STEP',
            update=update
            )
    user_defined_subs_enable : BoolProperty(
            name="User",
            update=update,
            default=True
            )
    user_defined_subs : StringProperty(
            name="User defined",
            update=update
            )
    idmat_subs : EnumProperty(
            name="Subs",
            items=materials_enum,
            default='1',
            update=update
            )
    panel : BoolProperty(
            name='Enable',
            default=True,
            update=update
            )
    panel_alt : FloatProperty(
            name="Altitude",
            default=0.25, precision=2, step=1,
            unit='LENGTH', subtype='DISTANCE',
            update=update
            )
    panel_x : FloatProperty(
            name="Width",
            min=0.001,
            default=0.01, precision=2, step=1,
            unit='LENGTH', subtype='DISTANCE',
            update=update
            )
    panel_z : FloatProperty(
            name="Height",
            min=0.001,
            default=0.6, precision=2, step=1,
            unit='LENGTH', subtype='DISTANCE',
            update=update
            )
    panel_dist : FloatProperty(
            name="Spacing",
            min=0.001,
            default=0.05, precision=2, step=1,
            unit='LENGTH', subtype='DISTANCE',
            update=update
            )
    panel_offset_x : FloatProperty(
            name="Offset",
            default=0.0, precision=2, step=1,
            unit='LENGTH', subtype='DISTANCE',
            update=update
            )
    idmat_panel : EnumProperty(
            name="Panels",
            items=materials_enum,
            default='2',
            update=update
            )
    rail : BoolProperty(
            name="Enable",
            update=update,
            default=False
            )
    rail_n : IntProperty(
            name="#",
            default=1,
            min=0,
            max=31,
            update=update
            )
    rail_x : FloatVectorProperty(
            name="Width",
            default=[
                0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05,
                0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05,
                0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05,
                0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05
            ],
            size=31,
            min=0.001,
            precision=2, step=1,
            unit='LENGTH',
            update=update
            )
    rail_z : FloatVectorProperty(
            name="Height",
            default=[
                0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05,
                0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05,
                0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05,
                0.05, 0.05, 0.05, 0.05, 0.05, 0.05, 0.05
            ],
            size=31,
            min=0.001,
            precision=2, step=1,
            unit='LENGTH',
            update=update
            )
    rail_offset : FloatVectorProperty(
            name="Offset",
            default=[
                0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 0, 0
            ],
            size=31,
            precision=2, step=1,
            unit='LENGTH',
            update=update
            )
    rail_alt : FloatVectorProperty(
            name="Altitude",
            default=[
                1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
                1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
                1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
                1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0
            ],
            size=31,
            precision=2, step=1,
            unit='LENGTH',
            update=update
            )
    rail_mat : CollectionProperty(type=archipack_fence_material)

    handrail : BoolProperty(
            name="Enable",
            update=update,
            default=True
            )
    handrail_offset : FloatProperty(
            name="Offset",
            default=0.0, precision=2, step=1,
            unit='LENGTH', subtype='DISTANCE',
            update=update
            )
    handrail_alt : FloatProperty(
            name="Altitude",
            default=1.0, precision=2, step=1,
            unit='LENGTH', subtype='DISTANCE',
            update=update
            )
    handrail_extend : FloatProperty(
            name="Extend",
            min=0,
            default=0.1, precision=2, step=1,
            unit='LENGTH', subtype='DISTANCE',
            update=update
            )
    handrail_slice : BoolProperty(
            name='Slice',
            default=True,
            update=update
            )
    handrail_slice_right : BoolProperty(
            name='Slice',
            default=True,
            update=update
            )
    handrail_profil : EnumProperty(
            name="Profil",
            items=(
                ('SQUARE', 'Square', '', 0),
                ('CIRCLE', 'Circle', '', 1),
                ('COMPLEX', 'Circle over square', '', 2)
                ),
            default='SQUARE',
            update=update
            )
    handrail_x : FloatProperty(
            name="Width",
            min=0.001,
            default=0.04, precision=2, step=1,
            unit='LENGTH', subtype='DISTANCE',
            update=update
            )
    handrail_y : FloatProperty(
            name="Height",
            min=0.001,
            default=0.04, precision=2, step=1,
            unit='LENGTH', subtype='DISTANCE',
            update=update
            )
    handrail_radius : FloatProperty(
            name="Radius",
            min=0.001,
            default=0.02, precision=2, step=1,
            unit='LENGTH', subtype='DISTANCE',
            update=update
            )
    idmat_handrail : EnumProperty(
            name="Handrail",
            items=materials_enum,
            default='0',
            update=update
            )

    # UI layout related
    parts_expand : BoolProperty(
            default=False
            )
    rail_expand : BoolProperty(
            default=False
            )
    idmats_expand : BoolProperty(
            default=False
            )
    handrail_expand : BoolProperty(
            default=False
            )
    post_expand : BoolProperty(
            default=False
            )
    panel_expand : BoolProperty(
            default=False
            )
    subs_expand : BoolProperty(
            default=False
            )

    # Flag to prevent mesh update while making bulk changes over variables
    # use :
    # .auto_update = False
    # bulk changes
    # .auto_update = True
    auto_update : BoolProperty(
            options={'SKIP_SAVE'},
            default=True,
            update=update_manipulators
            )

    def setup_manipulators(self):

        if len(self.manipulators) == 0:
            s = self.manipulators.add()
            s.prop1_name = "width"
            s = self.manipulators.add()
            s.prop1_name = "height"
            s.normal = Vector((0, 1, 0))

        for i in range(self.n_parts):
            p = self.parts[i]
            n_manips = len(p.manipulators)
            if n_manips == 0:
                s = p.manipulators.add()
                s.type_key = "ANGLE"
                s.prop1_name = "a0"
                s = p.manipulators.add()
                s.type_key = "SIZE"
                s.prop1_name = "length"
                s = p.manipulators.add()
                # s.type_key = 'SNAP_POINT'
                s.type_key = 'WALL_SNAP'
                s.prop1_name = str(i)
                s.prop2_name = 'post_z'

    def update_parts(self):

        # remove rails materials
        for i in range(len(self.rail_mat), self.rail_n, -1):
            self.rail_mat.remove(i - 1)

        # add rails
        for i in range(len(self.rail_mat), self.rail_n):
            self.rail_mat.add()

        # remove parts
        for i in range(len(self.parts), self.n_parts, -1):
            self.parts.remove(i - 1)

        # add parts
        for i in range(len(self.parts), self.n_parts):
            self.parts.add()
        self.setup_manipulators()

    def interpolate_bezier(self, pts, wM, p0, p1, resolution):
        # straight segment, worth testing here
        # since this can lower points count by a resolution factor
        # use normalized to handle non linear t
        if resolution == 0:
            pts.append(wM @ p0.co.to_3d())
        else:
            v = (p1.co - p0.co).normalized()
            d1 = (p0.handle_right - p0.co).normalized()
            d2 = (p1.co - p1.handle_left).normalized()
            if d1 == v and d2 == v:
                pts.append(wM @ p0.co.to_3d())
            else:
                seg = interpolate_bezier(wM @ p0.co,
                    wM @ p0.handle_right,
                    wM @ p1.handle_left,
                    wM @ p1.co,
                    resolution + 1)
                for i in range(resolution):
                    pts.append(seg[i].to_3d())

    def from_spline(self, context, wM, resolution, spline):

        o = self.find_in_selection(context)

        if o is None:
            return

        tM = wM.copy()
        tM.row[0].normalize()
        tM.row[1].normalize()
        tM.row[2].normalize()
        pts = []
        if spline.type == 'POLY':
            pt = spline.points[0].co
            pts = [wM @ p.co.to_3d() for p in spline.points]
            if spline.use_cyclic_u:
                pts.append(pts[0])
        elif spline.type == 'BEZIER':
            pt = spline.bezier_points[0].co
            points = spline.bezier_points
            for i in range(1, len(points)):
                p0 = points[i - 1]
                p1 = points[i]
                self.interpolate_bezier(pts, wM, p0, p1, resolution)
            if spline.use_cyclic_u:
                p0 = points[-1]
                p1 = points[0]
                self.interpolate_bezier(pts, wM, p0, p1, resolution)
                pts.append(pts[0])
            else:
                pts.append(wM @ points[-1].co)
        auto_update = self.auto_update
        self.auto_update = False

        self.n_parts = len(pts) - 1
        self.update_parts()

        p0 = pts.pop(0)
        a0 = 0
        for i, p1 in enumerate(pts):
            dp = p1 - p0
            da = atan2(dp.y, dp.x) - a0
            if da > pi:
                da -= 2 * pi
            if da < -pi:
                da += 2 * pi
            p = self.parts[i]
            p.length = dp.to_2d().length
            p.dz = dp.z
            p.a0 = da
            a0 += da
            p0 = p1

        self.auto_update = auto_update

        o.matrix_world = tM @ Matrix.Translation(pt)

    def update_path(self, context):
        path = context.scene.objects.get(self.user_defined_path.strip())
        if path is not None and path.type == 'CURVE':
            splines = path.data.splines
            if len(splines) > self.user_defined_spline:
                self.from_spline(
                    context,
                    path.matrix_world,
                    self.user_defined_resolution,
                    splines[self.user_defined_spline])

    def get_generator(self):
        g = FenceGenerator(self.parts)
        for part in self.parts:
            # type, radius, da, length
            g.add_part(part)

        g.set_offset(self.x_offset)
        # param_t(da, part_length)
        g.param_t(self.angle_limit, self.post_spacing)
        return g

    def update(self, context, manipulable_refresh=False):
        o = self.find_in_selection(context, self.auto_update)

        if o is None:
            return

        # clean up manipulators before any data model change
        if manipulable_refresh:
            self.manipulable_disable(context)

        self.update_parts()

        verts = []
        faces = []
        matids = []
        uvs = []

        g = self.get_generator()

        # depth at bottom
        # self.manipulators[1].set_pts([(0, 0, 0), (0, 0, self.height), (1, 0, 0)])

        if self.user_defined_post_enable:
            # user defined posts
            user_def_post = context.scene.objects.get(self.user_defined_post.strip())
            if user_def_post is not None and user_def_post.type == 'MESH':
                g.setup_user_defined_post(user_def_post, self.post_x, self.post_y, self.post_z)

        if self.post:
            g.make_post(0.5 * self.post_x, 0.5 * self.post_y, self.post_z,
                    self.post_alt, self.x_offset,
                    int(self.idmat_post), verts, faces, matids, uvs)

        # reset user def posts
        g.user_defined_post = None

        # user defined subs
        if self.user_defined_subs_enable:
            user_def_subs = context.scene.objects.get(self.user_defined_subs.strip())
            if user_def_subs is not None and user_def_subs.type == 'MESH':
                g.setup_user_defined_post(user_def_subs, self.subs_x, self.subs_y, self.subs_z)

        if self.subs:
            g.make_subs(0.5 * self.subs_x, 0.5 * self.subs_y, self.subs_z,
                    self.post_y, self.subs_alt, self.subs_spacing,
                    self.x_offset, self.subs_offset_x, int(self.idmat_subs), verts, faces, matids, uvs)

        g.user_defined_post = None

        if self.panel:
            g.make_panels(0.5 * self.panel_x, self.panel_z, self.post_y,
                    self.panel_alt, self.panel_dist, self.x_offset, self.panel_offset_x,
                    int(self.idmat_panel), verts, faces, matids, uvs)

        if self.rail:
            for i in range(self.rail_n):
                x = 0.5 * self.rail_x[i]
                y = self.rail_z[i]
                rail = [Vector((-x, y)), Vector((-x, 0)), Vector((x, 0)), Vector((x, y))]
                g.make_profile(rail, int(self.rail_mat[i].index), self.x_offset - self.rail_offset[i],
                        self.rail_alt[i], 0, verts, faces, matids, uvs)

        if self.handrail_profil == 'COMPLEX':
            sx = self.handrail_x
            sy = self.handrail_y
            handrail = [Vector((sx * x, sy * y)) for x, y in [
            (-0.28, 1.83), (-0.355, 1.77), (-0.415, 1.695), (-0.46, 1.605), (-0.49, 1.51), (-0.5, 1.415),
            (-0.49, 1.315), (-0.46, 1.225), (-0.415, 1.135), (-0.355, 1.06), (-0.28, 1.0), (-0.255, 0.925),
            (-0.33, 0.855), (-0.5, 0.855), (-0.5, 0.0), (0.5, 0.0), (0.5, 0.855), (0.33, 0.855), (0.255, 0.925),
            (0.28, 1.0), (0.355, 1.06), (0.415, 1.135), (0.46, 1.225), (0.49, 1.315), (0.5, 1.415),
            (0.49, 1.51), (0.46, 1.605), (0.415, 1.695), (0.355, 1.77), (0.28, 1.83), (0.19, 1.875),
            (0.1, 1.905), (0.0, 1.915), (-0.095, 1.905), (-0.19, 1.875)]]

        elif self.handrail_profil == 'SQUARE':
            x = 0.5 * self.handrail_x
            y = self.handrail_y
            handrail = [Vector((-x, y)), Vector((-x, 0)), Vector((x, 0)), Vector((x, y))]
        elif self.handrail_profil == 'CIRCLE':
            r = self.handrail_radius
            handrail = [Vector((r * sin(0.1 * -a * pi), r * (0.5 + cos(0.1 * -a * pi)))) for a in range(0, 20)]

        if self.handrail:
            g.make_profile(handrail, int(self.idmat_handrail), self.x_offset - self.handrail_offset,
                self.handrail_alt, self.handrail_extend, verts, faces, matids, uvs)

        bmed.buildmesh(context, o, verts, faces, matids=matids, uvs=uvs, weld=True, clean=True)

        # enable manipulators rebuild
        if manipulable_refresh:
            self.manipulable_refresh = True

        # restore context
        self.restore_context(context)

    def manipulable_setup(self, context):
        """
            NOTE:
            this one assume context.active_object is the instance this
            data belongs to, failing to do so will result in wrong
            manipulators set on active object
        """
        self.manipulable_disable(context)

        o = context.active_object

        self.setup_manipulators()

        for i, part in enumerate(self.parts):
            if i >= self.n_parts:
                break

            if i > 0:
                # start angle
                self.manip_stack.append(part.manipulators[0].setup(context, o, part))

            # length / radius + angle
            self.manip_stack.append(part.manipulators[1].setup(context, o, part))

            # snap point
            self.manip_stack.append(part.manipulators[2].setup(context, o, self))

        for m in self.manipulators:
            self.manip_stack.append(m.setup(context, o, self))


class ARCHIPACK_PT_fence(Panel):
    bl_idname = "ARCHIPACK_PT_fence"
    bl_label = "Fence"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = 'Archipack'

    @classmethod
    def poll(cls, context):
        return archipack_fence.filter(context.active_object)

    def draw(self, context):
        prop = archipack_fence.datablock(context.active_object)
        if prop is None:
            return
        scene = context.scene
        layout = self.layout
        row = layout.row(align=True)
        row.operator('archipack.fence_manipulate', icon='VIEW_PAN')
        box = layout.box()
        # box.label(text="Styles")
        row = box.row(align=True)
        row.operator("archipack.fence_preset_menu", text=bpy.types.ARCHIPACK_OT_fence_preset_menu.bl_label)
        row.operator("archipack.fence_preset", text="", icon='ADD')
        row.operator("archipack.fence_preset", text="", icon='REMOVE').remove_active = True
        box = layout.box()
        row = box.row(align=True)
        row.operator("archipack.fence_curve_update", text="", icon='FILE_REFRESH')
        row.prop_search(prop, "user_defined_path", scene, "objects", text="", icon='OUTLINER_OB_CURVE')
        if prop.user_defined_path != "":
            box.prop(prop, 'user_defined_spline')
            box.prop(prop, 'user_defined_resolution')
        box.prop(prop, 'angle_limit')
        box.prop(prop, 'x_offset')
        box = layout.box()
        row = box.row()
        if prop.parts_expand:
            row.prop(prop, 'parts_expand', icon="TRIA_DOWN", text="Parts", emboss=False)
            box.prop(prop, 'n_parts')
            for i, part in enumerate(prop.parts):
                part.draw(layout, context, i)
        else:
            row.prop(prop, 'parts_expand', icon="TRIA_RIGHT", text="Parts", emboss=False)

        box = layout.box()
        row = box.row(align=True)
        if prop.handrail_expand:
            row.prop(prop, 'handrail_expand', icon="TRIA_DOWN", text="Handrail", emboss=False)
        else:
            row.prop(prop, 'handrail_expand', icon="TRIA_RIGHT", text="Handrail", emboss=False)

        row.prop(prop, 'handrail')

        if prop.handrail_expand:
            box.prop(prop, 'handrail_alt')
            box.prop(prop, 'handrail_offset')
            box.prop(prop, 'handrail_extend')
            box.prop(prop, 'handrail_profil')
            if prop.handrail_profil != 'CIRCLE':
                box.prop(prop, 'handrail_x')
                box.prop(prop, 'handrail_y')
            else:
                box.prop(prop, 'handrail_radius')
            row = box.row(align=True)
            row.prop(prop, 'handrail_slice')

        box = layout.box()
        row = box.row(align=True)
        if prop.post_expand:
            row.prop(prop, 'post_expand', icon="TRIA_DOWN", text="Post", emboss=False)
        else:
            row.prop(prop, 'post_expand', icon="TRIA_RIGHT", text="Post", emboss=False)
        row.prop(prop, 'post')
        if prop.post_expand:
            box.prop(prop, 'post_spacing')
            box.prop(prop, 'post_x')
            box.prop(prop, 'post_y')
            box.prop(prop, 'post_z')
            box.prop(prop, 'post_alt')
            row = box.row(align=True)
            row.prop(prop, 'user_defined_post_enable', text="")
            row.prop_search(prop, "user_defined_post", scene, "objects", text="")

        box = layout.box()
        row = box.row(align=True)
        if prop.subs_expand:
            row.prop(prop, 'subs_expand', icon="TRIA_DOWN", text="Subs", emboss=False)
        else:
            row.prop(prop, 'subs_expand', icon="TRIA_RIGHT", text="Subs", emboss=False)

        row.prop(prop, 'subs')
        if prop.subs_expand:
            box.prop(prop, 'subs_spacing')
            box.prop(prop, 'subs_x')
            box.prop(prop, 'subs_y')
            box.prop(prop, 'subs_z')
            box.prop(prop, 'subs_alt')
            box.prop(prop, 'subs_offset_x')
            row = box.row(align=True)
            row.prop(prop, 'user_defined_subs_enable', text="")
            row.prop_search(prop, "user_defined_subs", scene, "objects", text="")

        box = layout.box()
        row = box.row(align=True)
        if prop.panel_expand:
            row.prop(prop, 'panel_expand', icon="TRIA_DOWN", text="Panels", emboss=False)
        else:
            row.prop(prop, 'panel_expand', icon="TRIA_RIGHT", text="Panels", emboss=False)
        row.prop(prop, 'panel')
        if prop.panel_expand:
            box.prop(prop, 'panel_dist')
            box.prop(prop, 'panel_x')
            box.prop(prop, 'panel_z')
            box.prop(prop, 'panel_alt')
            box.prop(prop, 'panel_offset_x')

        box = layout.box()
        row = box.row(align=True)
        if prop.rail_expand:
            row.prop(prop, 'rail_expand', icon="TRIA_DOWN", text="Rails", emboss=False)
        else:
            row.prop(prop, 'rail_expand', icon="TRIA_RIGHT", text="Rails", emboss=False)
        row.prop(prop, 'rail')
        if prop.rail_expand:
            box.prop(prop, 'rail_n')
            for i in range(prop.rail_n):
                box = layout.box()
                box.label(text="Rail " + str(i + 1))
                box.prop(prop, 'rail_x', index=i)
                box.prop(prop, 'rail_z', index=i)
                box.prop(prop, 'rail_alt', index=i)
                box.prop(prop, 'rail_offset', index=i)
                box.prop(prop.rail_mat[i], 'index', text="")

        box = layout.box()
        row = box.row()

        if prop.idmats_expand:
            row.prop(prop, 'idmats_expand', icon="TRIA_DOWN", text="Materials", emboss=False)
            box.prop(prop, 'idmat_handrail')
            box.prop(prop, 'idmat_panel')
            box.prop(prop, 'idmat_post')
            box.prop(prop, 'idmat_subs')
        else:
            row.prop(prop, 'idmats_expand', icon="TRIA_RIGHT", text="Materials", emboss=False)

# ------------------------------------------------------------------
# Define operator class to create object
# ------------------------------------------------------------------


class ARCHIPACK_OT_fence(ArchipackCreateTool, Operator):
    bl_idname = "archipack.fence"
    bl_label = "Fence"
    bl_description = "Fence"
    bl_category = 'Archipack'
    bl_options = {'REGISTER', 'UNDO'}

    def create(self, context):

        m = bpy.data.meshes.new("Fence")
        o = bpy.data.objects.new("Fence", m)
        d = m.archipack_fence.add()
        # make manipulators selectable

        d.manipulable_selectable = True
        self.link_object_to_scene(context, o)
        o.select_set(state=True)
        context.view_layer.objects.active = o
        self.load_preset(d)

        self.add_material(o)
        return o

    def execute(self, context):
        if context.mode == "OBJECT":
            bpy.ops.object.select_all(action="DESELECT")
            o = self.create(context)
            o.location = context.scene.cursor.location
            o.select_set(state=True)
            context.view_layer.objects.active = o
            self.manipulate()
            return {'FINISHED'}
        else:
            self.report({'WARNING'}, "Archipack: Option only valid in Object mode")
            return {'CANCELLED'}


# ------------------------------------------------------------------
# Define operator class to create object
# ------------------------------------------------------------------

class ARCHIPACK_OT_fence_curve_update(Operator):
    bl_idname = "archipack.fence_curve_update"
    bl_label = "Fence curve update"
    bl_description = "Update fence data from curve"
    bl_category = 'Archipack'
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(self, context):
        return archipack_fence.filter(context.active_object)

    def draw(self, context):
        layout = self.layout
        row = layout.row()
        row.label(text="Use Properties panel (N) to define parms", icon='INFO')

    def execute(self, context):
        if context.mode == "OBJECT":
            d = archipack_fence.datablock(context.active_object)
            d.update_path(context)
            return {'FINISHED'}
        else:
            self.report({'WARNING'}, "Archipack: Option only valid in Object mode")
            return {'CANCELLED'}


class ARCHIPACK_OT_fence_from_curve(ArchipackCreateTool, Operator):
    bl_idname = "archipack.fence_from_curve"
    bl_label = "Fence curve"
    bl_description = "Create a fence from a curve"
    bl_category = 'Archipack'
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(self, context):
        return context.active_object is not None and context.active_object.type == 'CURVE'

    def draw(self, context):
        layout = self.layout
        row = layout.row()
        row.label(text="Use Properties panel (N) to define parms", icon='INFO')

    def create(self, context):
        o = None
        curve = context.active_object
        for i, spline in enumerate(curve.data.splines):
            bpy.ops.archipack.fence('INVOKE_DEFAULT', auto_manipulate=False)
            o = context.active_object
            d = archipack_fence.datablock(o)
            d.auto_update = False
            d.user_defined_spline = i
            d.user_defined_path = curve.name
            d.auto_update = True
        return o

    def execute(self, context):
        if context.mode == "OBJECT":
            bpy.ops.object.select_all(action="DESELECT")
            o = self.create(context)
            if o is not None:
                o.select_set(state=True)
                context.view_layer.objects.active = o
            # self.manipulate()
            return {'FINISHED'}
        else:
            self.report({'WARNING'}, "Archipack: Option only valid in Object mode")
            return {'CANCELLED'}

# ------------------------------------------------------------------
# Define operator class to manipulate object
# ------------------------------------------------------------------


class ARCHIPACK_OT_fence_manipulate(Operator):
    bl_idname = "archipack.fence_manipulate"
    bl_label = "Manipulate"
    bl_description = "Manipulate"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(self, context):
        return archipack_fence.filter(context.active_object)

    def invoke(self, context, event):
        d = archipack_fence.datablock(context.active_object)
        d.manipulable_invoke(context)
        return {'FINISHED'}


# ------------------------------------------------------------------
# Define operator class to load / save presets
# ------------------------------------------------------------------


class ARCHIPACK_OT_fence_preset_menu(PresetMenuOperator, Operator):
    bl_description = "Show Fence Presets"
    bl_idname = "archipack.fence_preset_menu"
    bl_label = "Fence Styles"
    preset_subdir = "archipack_fence"


class ARCHIPACK_OT_fence_preset(ArchipackPreset, Operator):
    """Add a Fence Preset"""
    bl_idname = "archipack.fence_preset"
    bl_label = "Add Fence Style"
    preset_menu = "ARCHIPACK_OT_fence_preset_menu"

    @property
    def blacklist(self):
        return ['manipulators', 'n_parts', 'parts', 'user_defined_path', 'user_defined_spline']


def register():
    bpy.utils.register_class(archipack_fence_material)
    bpy.utils.register_class(archipack_fence_part)
    bpy.utils.register_class(archipack_fence)
    Mesh.archipack_fence = CollectionProperty(type=archipack_fence)
    bpy.utils.register_class(ARCHIPACK_OT_fence_preset_menu)
    bpy.utils.register_class(ARCHIPACK_PT_fence)
    bpy.utils.register_class(ARCHIPACK_OT_fence)
    bpy.utils.register_class(ARCHIPACK_OT_fence_preset)
    bpy.utils.register_class(ARCHIPACK_OT_fence_manipulate)
    bpy.utils.register_class(ARCHIPACK_OT_fence_from_curve)
    bpy.utils.register_class(ARCHIPACK_OT_fence_curve_update)


def unregister():
    bpy.utils.unregister_class(archipack_fence_material)
    bpy.utils.unregister_class(archipack_fence_part)
    bpy.utils.unregister_class(archipack_fence)
    del Mesh.archipack_fence
    bpy.utils.unregister_class(ARCHIPACK_OT_fence_preset_menu)
    bpy.utils.unregister_class(ARCHIPACK_PT_fence)
    bpy.utils.unregister_class(ARCHIPACK_OT_fence)
    bpy.utils.unregister_class(ARCHIPACK_OT_fence_preset)
    bpy.utils.unregister_class(ARCHIPACK_OT_fence_manipulate)
    bpy.utils.unregister_class(ARCHIPACK_OT_fence_from_curve)
    bpy.utils.unregister_class(ARCHIPACK_OT_fence_curve_update)



# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

# <pep8 compliant>


import bpy
import bmesh
from mathutils.geometry import intersect_line_line as LineIntersect

import itertools
from collections import defaultdict
from . import cad_module as cm


def order_points(edge, point_list):
    ''' order these edges from distance to v1, then
    sandwich the sorted list with v1, v2 '''
    v1, v2 = edge

    def dist(co):
        return (v1 - co).length
    point_list = sorted(point_list, key=dist)
    return [v1] + point_list + [v2]


def remove_permutations_that_share_a_vertex(bm, permutations):
    ''' Get useful Permutations '''
    final_permutations = []
    for edges in permutations:
        raw_vert_indices = cm.vertex_indices_from_edges_tuple(bm, edges)
        if cm.duplicates(raw_vert_indices):
            continue

        # reaches this point if they do not share.
        final_permutations.append(edges)

    return final_permutations


def get_valid_permutations(bm, edge_indices):
    raw_permutations = itertools.permutations(edge_indices, 2)
    permutations = [r for r in raw_permutations if r[0] < r[1]]
    return remove_permutations_that_share_a_vertex(bm, permutations)


def can_skip(closest_points, vert_vectors):
    '''this checks if the intersection lies on both edges, returns True
    when criteria are not met, and thus this point can be skipped'''
    if not closest_points:
        return True
    if not isinstance(closest_points[0].x, float):
        return True
    if cm.num_edges_point_lies_on(closest_points[0], vert_vectors) < 2:
        return True

    # if this distance is larger than than VTX_PRECISION, we can skip it.
    cpa, cpb = closest_points
    return (cpa - cpb).length > cm.CAD_prefs.VTX_PRECISION


def get_intersection_dictionary(bm, edge_indices):

    bm.verts.ensure_lookup_table()
    bm.edges.ensure_lookup_table()

    permutations = get_valid_permutations(bm, edge_indices)

    k = defaultdict(list)
    d = defaultdict(list)

    for edges in permutations:
        raw_vert_indices = cm.vertex_indices_from_edges_tuple(bm, edges)
        vert_vectors = cm.vectors_from_indices(bm, raw_vert_indices)

        points = LineIntersect(*vert_vectors)

        # some can be skipped.    (NaN, None, not on both edges)
        if can_skip(points, vert_vectors):
            continue

        # reaches this point only when an intersection happens on both edges.
        [k[edge].append(points[0]) for edge in edges]

    # k will contain a dict of edge indices and points found on those edges.
    for edge_idx, unordered_points in k.items():
        tv1, tv2 = bm.edges[edge_idx].verts
        v1 = bm.verts[tv1.index].co
        v2 = bm.verts[tv2.index].co
        ordered_points = order_points((v1, v2), unordered_points)
        d[edge_idx].extend(ordered_points)

    return d


def update_mesh(bm, d):
    ''' Make new geometry (delete old first) '''

    oe = bm.edges
    ov = bm.verts

    new_verts = []
    collect = new_verts.extend
    for old_edge, point_list in d.items():
        num_edges_to_add = len(point_list)-1
        for i in range(num_edges_to_add):
            a = ov.new(point_list[i])
            b = ov.new(point_list[i+1])
            oe.new((a, b))
            bm.normal_update()
            collect([a, b])

    bmesh.ops.delete(bm, geom=[edge for edge in bm.edges if edge.select], context='EDGES')

    #bpy.ops.mesh.remove_doubles(
    #    threshold=cm.CAD_prefs.VTX_DOUBLES_THRSHLD,
    #    use_unselected=False)

    bmesh.ops.remove_doubles(bm, verts=new_verts, dist=cm.CAD_prefs.VTX_DOUBLES_THRSHLD)


def unselect_nonintersecting(bm, d_edges, edge_indices):
    if len(edge_indices) > len(d_edges):
        reserved_edges = set(edge_indices) - set(d_edges)
        for edge in reserved_edges:
            bm.edges[edge].select = False
        print("unselected {}, non intersecting edges".format(reserved_edges))


class TCIntersectAllEdges(bpy.types.Operator):
    '''Adds a vertex at the intersections of all selected edges'''
    bl_idname = 'tinycad.intersectall'
    bl_label = 'XALL intersect all edges'
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        obj = context.active_object
        return obj is not None and obj.type == 'MESH' and obj.mode == 'EDIT'

    def execute(self, context):
        # must force edge selection mode here
        bpy.context.tool_settings.mesh_select_mode = (False, True, False)

        obj = context.active_object
        if obj.mode == "EDIT":
            bm = bmesh.from_edit_mesh(obj.data)

            selected_edges = [edge for edge in bm.edges if edge.select]
            edge_indices = [i.index for i in selected_edges]

            d = get_intersection_dictionary(bm, edge_indices)

            unselect_nonintersecting(bm, d.keys(), edge_indices)
            update_mesh(bm, d)

            bmesh.update_edit_mesh(obj.data)
        else:
            print('must be in edit mode')

        return {'FINISHED'}



# Copyright 2018-2021 The glTF-Blender-IO authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing
import array
from io_scene_gltf2.io.com import gltf2_io_constants


class BinaryData:
    """Store for gltf binary data that can later be stored in a buffer."""

    def __init__(self, data: bytes):
        if not isinstance(data, bytes):
            raise TypeError("Data is not a bytes array")
        self.data = data

    def __eq__(self, other):
        return self.data == other.data

    def __hash__(self):
        return hash(self.data)

    @classmethod
    def from_list(cls, lst: typing.List[typing.Any], gltf_component_type: gltf2_io_constants.ComponentType):
        format_char = gltf2_io_constants.ComponentType.to_type_code(gltf_component_type)
        return BinaryData(array.array(format_char, lst).tobytes())

    @property
    def byte_length(self):
        return len(self.data)



import bpy
material = (bpy.context.material.active_node_material if bpy.context.material.active_node_material else bpy.context.material)

material.subsurface_scattering.radius = 14.266, 7.228, 2.036
material.subsurface_scattering.color = 0.855, 0.740, 0.292




import os
PATHS = (
    "doc",
    "release/scripts/modules",
    "release/scripts/startup",
    "release/scripts/templates_py",
    "source/blender",
    "source/tools",
    "tests",
)

BLACKLIST = (
    "source/tools/svn_rev_map/sha1_to_rev.py",
    "source/tools/svn_rev_map/rev_to_sha1.py",
    "source/tools/svn_rev_map/rev_to_sha1.py",
    "release/scripts/modules/bl_i18n_utils/settings.py",
    "release/scripts/modules/bl_i18n_utils/utils_rtl.py",
)

SOURCE_DIR = os.path.normpath(os.path.abspath(os.path.normpath(
    os.path.join(os.path.dirname(__file__), "..", "..", ".."))))

PATHS = tuple(
    os.path.join(SOURCE_DIR, p.replace("/", os.sep))
    for p in PATHS
)

BLACKLIST = set(
    os.path.join(SOURCE_DIR, p.replace("/", os.sep))
    for p in BLACKLIST
)


def files(path, test_fn):
    for dirpath, dirnames, filenames in os.walk(path):
        # skip '.git'
        dirnames[:] = [d for d in dirnames if not d.startswith(".")]
        for filename in filenames:
            if test_fn(filename):
                filepath = os.path.join(dirpath, filename)
                yield filepath
